package com.ichaoj.ams.service.impl;

import cn.hutool.core.bean.BeanUtil;
import cn.hutool.core.collection.CollectionUtil;
import cn.hutool.core.thread.ThreadUtil;
import cn.hutool.core.util.StrUtil;
import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.baomidou.mybatisplus.core.toolkit.Wrappers;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.baomidou.mybatisplus.extension.service.impl.ServiceImpl;
import com.ichaoj.ams.entity.*;
import com.ichaoj.ams.mapper.AmsExecuteRecordMapper;
import com.ichaoj.ams.request.execute.CreateExecute;
import com.ichaoj.ams.request.execute.PageExecuteRequest;
import com.ichaoj.ams.request.execute.UpdateExecute;
import com.ichaoj.ams.response.execute.ExecuteResponse;
import com.ichaoj.ams.service.*;
import com.ichaoj.common.exception.ErrorServiceException;
import com.ichaoj.common.model.PublicPage;
import com.ichaoj.common.model.PublicUserInfo;
import com.ichaoj.mybatis.service.SuperWhaleServiceImpl;
import com.ichaoj.web.context.SuperWhaleContext;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import javax.annotation.Resource;
import java.time.LocalDateTime;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;

/**
 * <p>
 * 自动化 服务实现类
 * </p>
 *
 * @author cjwen
 * @since 2023-05-18
 */
@Service
public class AmsExecuteRecordServiceImpl extends SuperWhaleServiceImpl<AmsExecuteRecordMapper, AmsExecuteRecord> implements IAmsExecuteRecordService {
    @Resource
    private IAmsAirdropTaskService taskService;

    @Resource
    private IAmsAirdropProjectService projectService;

    @Resource
    private IAmsAddressAccountService accountService;

    @Resource
    private IAmsTradeRecordService tradeService;

    @Override
    public PublicPage<ExecuteResponse> pageExecute(PageExecuteRequest executeRequest) {
        LambdaQueryWrapper<AmsExecuteRecord> wrapper = Wrappers.lambdaQuery();
        wrapper
                .eq(StrUtil.isNotBlank(executeRequest.getQueryValue()), AmsExecuteRecord::getExecuteId, executeRequest.getQueryValue())
                .or(w -> w.like(StrUtil.isNotBlank(executeRequest.getQueryValue()), AmsExecuteRecord::getGroupName, executeRequest.getQueryValue())
                        .or().like(StrUtil.isNotBlank(executeRequest.getQueryValue()), AmsExecuteRecord::getTaskId, executeRequest.getQueryValue()))
                .orderByDesc(AmsExecuteRecord::getCreateTime);
        Page<AmsExecuteRecord> result = this.page(this.buildPageObj(executeRequest), wrapper);
        return this.convertPublicPage(result, s -> BeanUtil.copyProperties(s, ExecuteResponse.class));
    }

    @Override
    @Transactional(rollbackFor = Exception.class)
    public void createExecute(CreateExecute createExecute) {
        String userId = SuperWhaleContext.getContext(PublicUserInfo.class).getUserId();
        if (StrUtil.isBlank(createExecute.getGroupName())) {
            throw new ErrorServiceException("地址组名称不能为空！");
        }
        List<AmsAddressAccount> accountList = accountService.getByGroupNameAndUserId(createExecute.getGroupName(), userId);
        if (CollectionUtil.isEmpty(accountList)) {
            throw new ErrorServiceException("地址组错误！");
        }
        if (accountList.get(0).getAddressType() != 1) {
            throw new ErrorServiceException(createExecute.getGroupName() + "为非批量号地址组，不允许交易");
        }
        AmsAirdropProject project = verifyProject(createExecute.getProjectId());
        AmsAirdropTask task = verifyTask(createExecute.getTaskId(), project.getAmsProjectId());
        AmsExecuteRecord amsExecute = BeanUtil.copyProperties(createExecute, AmsExecuteRecord.class);
        amsExecute.setUserId(userId);
        amsExecute.setCreateTime(LocalDateTime.now());
        amsExecute.setTaskId(task.getAmsTaskId());
        amsExecute.setProjectId(project.getAmsProjectId());
        // todo 查询是第几次执行
        this.save(amsExecute);
        if (StrUtil.isBlank(createExecute.getPassword())) {
            throw new ErrorServiceException("密码不能为空！");
        }
        // todo 验证密码是否正确

        ThreadUtil.execute(() -> tradeService.randomTrans(accountList
                , createExecute.getIntervalMin()
                , createExecute.getIntervalMax()
                , createExecute.getAmount()
                , createExecute.getMaxGas()
                , amsExecute.getExecuteId()));

    }

    @Override
    public void updateExecute(UpdateExecute updateExecute) {
        AmsAirdropProject project = verifyProject(updateExecute.getProjectId());
        AmsAirdropTask task = verifyTask(updateExecute.getTaskId(), project.getAmsProjectId());
        AmsExecuteRecord execute = BeanUtil.copyProperties(updateExecute, AmsExecuteRecord.class);
        execute.setUserId(SuperWhaleContext.getContext(PublicUserInfo.class).getUserId());
        execute.setUpdateTime(LocalDateTime.now());
        execute.setTaskId(task.getAmsTaskId());
        this.updateById(execute);
    }

    /**
     * 验证项目id
     *
     * @param projectId 项目id
     * @return 项目实体
     */
    private AmsAirdropProject verifyProject(String projectId) {
        if (StrUtil.isBlank(projectId)) {
            throw new ErrorServiceException("projectId is required!");
        }
        AmsAirdropProject project = projectService.getById(projectId);
        if (project == null) {
            throw new ErrorServiceException("Airdrop Project Error!");
        }
        return project;
    }

    /**
     * 验证任务id
     *
     * @param taskId    任务id
     * @param projectId 项目id
     * @return 任务实体
     */
    private AmsAirdropTask verifyTask(String taskId, String projectId) {
        if (StrUtil.isBlank(taskId)) {
            throw new ErrorServiceException("taskId is required!");
        }
        LambdaQueryWrapper<AmsAirdropTask> wrapper = Wrappers.lambdaQuery();
        wrapper
                .eq(AmsAirdropTask::getAirdropProjectId, projectId)
                .eq(AmsAirdropTask::getAmsTaskId, taskId);
        AmsAirdropTask task = taskService.getOne(wrapper);
        if (task == null) {
            throw new ErrorServiceException("Task Error!");
        }
        return task;
    }

    @Scheduled(cron = "0/59 * * * * ? ")
    public void scanExecuteStatus() {
        List<AmsExecuteRecord> list = this.list();
        for (AmsExecuteRecord executeRecord : list) {
            // 验证用户执行的状态
            String executeId = executeRecord.getExecuteId();
            String userId = executeRecord.getUserId();
            String groupName = executeRecord.getGroupName();
            List<AmsAddressAccount> accounts = accountService.getByGroupNameAndUserId(groupName, userId);
            Set<String> predictSet = accounts.stream().map(AmsAddressAccount::getAddress).collect(Collectors.toSet());
            List<AmsTradeRecord> tradeRecords = tradeService.list(new LambdaQueryWrapper<AmsTradeRecord>()
                    .eq(AmsTradeRecord::getExecuteId, executeId));
            Set<String> actualSet = tradeRecords.stream().map(AmsTradeRecord::getAddress).collect(Collectors.toSet());
            // 判断交易记录的地址是否
            if (CollectionUtil.containsAll(predictSet, actualSet) && CollectionUtil.containsAll(actualSet, predictSet)) {
                Set<String> confirmSet = new HashSet<>();
                for (AmsTradeRecord tradeRecord : tradeRecords) {
                    if (tradeRecord.getStatus() == 1) {
                        confirmSet.add(tradeRecord.getAddress());
                    }
                }
                // 修改状态
                if (CollectionUtil.containsAll(confirmSet, actualSet) && CollectionUtil.containsAll(actualSet, confirmSet)) {
                    executeRecord.setExecuteStatus(1);
                    this.updateById(executeRecord);
                }
            }
        }
    }
}
